from __future__ import annotations

from typing import Callable, Dict, List, Tuple

import torch as tt
from PIL import Image

from rex_xai.explanation.explanation import Explanation
from rex_xai.explanation.rex import calculate_responsibility, predict_target
from rex_xai.input.config import CausalArgs
from rex_xai.input.input_data import Data
from rex_xai.output import visualisation
from rex_xai.responsibility.prediction import default_prediction_function
from rex_xai.responsibility.resp_maps import ResponsibilityMaps
from rex_xai.utils._utils import Strategy


class ReX:
    def __init__(
        self,
        model,
        model_shape: List[int | str] | Tuple[int | str],
        device: str | tt.device,
        mode: str,
        prediction_function: Callable | None = None,
    ) -> None:
        self.args: CausalArgs = CausalArgs()
        self.model = model
        self.model_shape: List[int | str] | Tuple[int | str] = model_shape
        self.device: str | tt.device = device
        self.mode: str = mode
        self.explanation = None
        self.prediction_function = prediction_function
        self.data = None
        self.maps: ResponsibilityMaps | None = None
        self.stats: Dict[str, float] | None = None

        if self.prediction_function is None:
            self.get_default_prediction_function()

    def set_tabular_data(self, path: str, data=None):
        if self.mode == "spectral":
            self.args.path = path
            self.args.mode = "spectral"
            if data is not None:
                self.data = Data(
                    data,
                    data.shape,
                    self.device,
                    mode=self.args.mode,
                    process=True,
                )

    def set_input(self, input):
        if self.data is not None:
            self.data.input = input

    def set_transformed_data(self, transformed_data: tt.Tensor):
        if self.data is not None:
            self.data.data = transformed_data

    def set_rgb_image(self, path: str):
        if self.mode == "RGB":
            self.args.path = path
            self.args.mode = "RGB"
            img = Image.open(self.args.path).convert("RGB")
            self.data = Data(img, self.model_shape, self.device, mode="RGB")
            return img

    def show_target(self):
        if self.data is not None and self.data.target is not None:
            print(self.data.target)
        else:
            print("a target has not yet been set")

    def get_default_prediction_function(self):
        self.prediction_function = default_prediction_function(self.model)

    def set_target(self):
        if self.data is not None:
            self.data.target = predict_target(
                self.data, self.args, self.prediction_function
            )
        return self

    def set_prediction_function(self, function):
        self.prediction_function = function

    def calculate_responsibility(self, args: CausalArgs | None = None):
        if args is None:
            args = self.args
        if self.data is not None:
            if self.data.mask_value is None:
                self.data.set_mask_value(args.mask_value)
            maps, stats = calculate_responsibility(
                self.data, args, self.prediction_function
            )
            self.maps = maps
            self.stats = stats
            return self

    def generate_explanation_object(self):
        if self.data is not None and self.maps is not None and self.stats is not None:
            self.explanation = Explanation(
                self.maps,
                self.prediction_function,
                self.data,
                self.args,
                self.stats,
            )
        return self

    def extract_sufficient_explanation(self):
        if self.explanation is not None:
            self.explanation.extract()
        else:
            self.generate_explanation_object().explanation.extract()  # type: ignore

    def extract_contrastive_explanation(self):
        self.args.strategy = Strategy.Contrastive
        self.args.complete = False
        self.generate_explanation_object().explanation.extract()  # type: ignore

    def extract_complete_explanation(self):
        self.args.strategy = Strategy.Contrastive
        self.args.complete = True
        self.generate_explanation_object().explanation.extract()  # type: ignore

    def analyse(self):
        pass

    def rerun_with(self, new_args: CausalArgs):
        self.calculate_responsibility(args=new_args)

    def show(self):
        if self.explanation is not None and self.data is not None:
            if self.mode == "RGB":
                if self.args.complete:
                    return visualisation.save_complete(
                        self.explanation, self.data, self.args
                    )
                if self.args.strategy == Strategy.Contrastive:
                    return visualisation.save_contrastive(
                        self.explanation, self.data, self.args
                    )
                elif self.args.strategy == Strategy.Global:
                    return visualisation.save_image(
                        self.explanation.sufficiency_mask,  # type: ignore
                        self.data,
                        self.args,  # type: ignore
                    )
                else:
                    pass

            if self.mode == "spectral":
                visualisation.spectral_plot(
                    self.explanation.sufficiency_mask,
                    self.data,  # type: ignore
                    self.maps.get(self.data.target.classification),  # type: ignore
                    self.args.heatmap_colours,
                )



"""distributions module"""

from typing import Optional, Tuple
from enum import Enum
import numpy as np
from scipy.stats import betabinom
from rex_xai.utils.logger import logger

Distribution = Enum("Distribution", ["Binomial", "Uniform", "BetaBinomial", "Adaptive"])


def _betabinom2d(height, width, alpha, beta):
    bx = betabinom(width, alpha, beta)
    by = betabinom(height, alpha, beta)

    w = np.array([bx.pmf(i) for i in range(0, width)])  # type: ignore
    h = np.array([by.pmf(i) for i in range(0, height)])  # type: ignore

    w = np.expand_dims(w, axis=0)
    h = np.expand_dims(h, axis=0)

    u = (h.T * w / np.sum(h.T * w)).ravel()
    p = np.random.choice(np.arange(0, len(u)), p=u)
    return p


def _blend(dist, alpha, base):
    pass
    # def random_coords(d: Optional[Distribution], *args, map=None) -> Optional[int]:


#     pmf = np.array([base.pmf(x) for x in range(0, len(dist))])
#     blend = ((1.0 - alpha) * pmf) + (alpha * dist)
#     blend /= np.sum(blend)
#     return blend
#
#
def _2d_adaptive(map, args: Tuple[int, int, int, int], alpha=0.0, base=None) -> int:
    return 0


#     # if the map exists and is not 0.0 everywhere...
#     if map is not None and np.max(map) > 0.0:
#         s = map[args[0] : args[1], args[2] : args[3]]
#         sf = np.ndarray.flatten(s)
#         # sf = np.max(sf) - sf
#         sf /= np.sum(sf)
#
#         # base = betabinom(0, len(sf), 1.1, 1.1)
#         # if base is not None:
#         #     sf = _blend(alpha, base)
#         pos = np.random.choice(np.arange(0, len(sf)), p=sf)
#         return pos
#
#     # if the map is empty or doesn't exist, return uniform
#     return np.random.randint(1, (args[1] - args[0]) * (args[3] - args[2]))


def str2distribution(d: str) -> Distribution:
    """converts string into Distribution enum"""
    if d == "binom":
        return Distribution.Binomial
    elif d == "uniform":
        return Distribution.Uniform
    elif d == "betabinom":
        return Distribution.BetaBinomial
    elif d == "adaptive":
        return Distribution.Adaptive
    else:
        logger.warning(
            "Invalid distribution '%s', reverting to uniform distribution",
            d,
        )
    return Distribution.Uniform


def random_coords(d: Optional[Distribution], *args, map=None):
    """generates random coordinates given a distribution and args"""

    try:
        if d == Distribution.Adaptive:
            return _2d_adaptive(map, args[0])

        if d == Distribution.Uniform:
            if args[1] == 1:
                return np.random.randint(1, args[0])
            else:
                return np.random.choice(args[0], args[1], replace=False)

        if d == Distribution.BetaBinomial:
            if args[1] == 1:
                return _betabinom2d(args[3], args[4], args[2][0], args[2][1])
            elif args[1] == 3:
                y = betabinom(args[0], *args[2])
                pmf = np.array([y.pmf(i) for i in range(0, args[0] + 1)])  # type:ignore
                return np.random.choice(args[0] + 1, 3, replace=False, p=pmf)
            else:
                pass

        return None
    except ValueError:
        return None



#!/usr/bin/env python
import numbers
from typing import List, Optional

import numpy as np
import torch as tt
from PIL import Image  # type: ignore

try:
    from anytree.cachedsearch import find
except ImportError:
    from anytree.search import find

import matplotlib.pyplot as plt

from rex_xai.input.input_data import Data
from rex_xai.mutants.box import Box
from rex_xai.responsibility.prediction import Prediction
from rex_xai.utils._utils import add_boundaries, set_boolean_mask_value, try_detach
from rex_xai.utils.logger import logger

__combinations = [
    [
        0,
    ],
    [
        1,
    ],
    [
        2,
    ],
    [
        3,
    ],
    [0, 1],
    [0, 2],
    [0, 3],
    [1, 2],
    [1, 3],
    [2, 3],
    [0, 1, 2],
    [0, 1, 3],
    [0, 2, 3],
    [1, 2, 3],
]


def _apply_to_data(mask, data: Data):
    if callable(data.mask_value):
        return data.mask_value(mask, data.data)
    if isinstance(data.mask_value, numbers.Number):
        return tt.where(mask, data.data, data.mask_value)  # type: ignore

    print(data.mask_value)
    logger.warning("applying default masking value of 0")
    return tt.where(mask, data.data, 0)  # type: ignore


def get_combinations():
    return __combinations


class Mutant:
    def __init__(self, data: Data, static, active, masking_func) -> None:
        self.shape = tuple(
            data.model_shape[1:]
        )  # the first element of shape is the batch information, so we drop that
        self.mode = data.mode
        self.channels: int = (
            data.model_channels if data.model_channels is not None else 1
        )
        self.order = data.model_order
        self.mask = tt.zeros(self.shape, dtype=tt.bool, device=data.device)
        self.static = static
        self.active = active
        self.prediction: Optional[Prediction] = None
        self.passing = False
        self.masking_func = masking_func
        self.depth = 0

    def __repr__(self) -> str:
        return f"ACTIVE: {self.active}, PREDICTION: {self.prediction}, PASSING: {self.passing}"

    def get_name(self):
        return self.active

    def update_status(self, target):
        if self.prediction is not None:
            if target.classification == self.prediction.classification:
                self.passing = True

    def get_length(self):
        return len(self.active.split("_"))

    def get_active_boxes(self):
        return self.active.split("_")

    def area(self) -> int:
        """Return the total area *not* concealed by the mutant."""
        tensor = tt.count_nonzero(self.mask)
        if tensor.numel() == 0 or tensor is None:
            return 0
        else:
            return int(tensor.item()) // self.channels

    def set_static_mask_regions(self, names, search_tree):
        for name in names:
            box = find(search_tree, lambda node: node.name == name)
            if box is not None:
                self.depth = max(self.depth, box.depth)
                self.set_mask_region_to_true(box)

    def set_active_mask_regions(self, boxes: List[Box]):
        for box in boxes:
            self.depth = max(self.depth, box.depth)
            self.set_mask_region_to_true(box)

    def set_mask_region_to_true(self, box: Box):
        set_boolean_mask_value(self.mask, self.mode, self.order, box)

    def apply_to_data(self, data: Data):
        return _apply_to_data(self.mask, data)

    def save_mutant(self, data: Data, name=None, segs=None):
        if data.mode == "RGB":
            m = np.array(data.input)
            mask = try_detach(self.mask).squeeze()

            if data.transposed:
                # if transposed, we have C * H * W, so change that to H * W * C
                m = np.where(mask, m.transpose((2, 0, 1)), 0)
                m = m.transpose((1, 2, 0))
            else:
                mask = mask.transpose((1, 2, 0))
                m = np.where(mask, m, 0)
            # draw on the segment_mask, if available
            if segs is not None:
                m = add_boundaries(m, segs)
            img = Image.fromarray(m, data.mode)
            if name is not None:
                img.save(name)
            else:
                img.save(f"{self.get_name()}.png")
        # spectral or time series data
        if data.mode == "spectral":
            m = self.apply_to_data(data)
            fig = plt.figure()
            ax = fig.add_subplot(111)
            ax.plot(m[0][0].detach().cpu().numpy())
            plt.savefig(f"{self.get_name()}.png")
        # 3d image
        if data.mode == "voxel":
            volume = self.apply_to_data(data).squeeze().detach().cpu().numpy()
            num_slices = min(volume.shape[0], 8)
            slice_indices_x = np.linspace(0, volume.shape[0] - 1, num_slices, dtype=int)
            slice_indices_y = np.linspace(0, volume.shape[1] - 1, num_slices, dtype=int)
            slice_indices_z = np.linspace(0, volume.shape[2] - 1, num_slices, dtype=int)

            fig, axes = plt.subplots(3, num_slices, figsize=(15, 6))
            for i, idx in enumerate(slice_indices_x):
                ax = axes[0, i]
                ax.imshow(volume[idx, :, :], cmap="gray")
                ax.set_title(f"X={idx}")
                ax.axis("off")

            for j, idy in enumerate(slice_indices_y):
                ax = axes[1, j]
                ax.imshow(volume[:, idy, :], cmap="gray")
                ax.set_title(f"Y={idy}")
                ax.axis("off")

            for z, idz in enumerate(slice_indices_z):
                ax = axes[2, z]
                ax.imshow(volume[:, :, idz], cmap="gray")
                ax.set_title(f"Z={idz}")
                ax.axis("off")

            plt.tight_layout()
            plt.savefig(name or f"{self.get_name()}.png")
            plt.close(fig)



#!/usr/bin/env python
from __future__ import annotations

import zlib
from ast import literal_eval
from datetime import datetime

import numpy as np
import pandas as pd
import sqlalchemy as sa
import torch as tt
from sqlalchemy import Boolean, Column, Float, Integer, String, Unicode, create_engine
from sqlalchemy.orm import DeclarativeBase, sessionmaker

from rex_xai.explanation.explanation import Explanation
from rex_xai.explanation.multi_explanation import MultiExplanation
from rex_xai.input.config import CausalArgs, Strategy
from rex_xai.utils._utils import try_detach
from rex_xai.utils.logger import logger


def _dataframe(db, table):
    return pd.read_sql_table(table, f"sqlite:///{db}")


def _to_numpy(buffer, shape, dtype):
    if buffer is None:
        return None
    return np.frombuffer(zlib.decompress(buffer), dtype=dtype).reshape(shape)


def process_responsibility_column(df, dtype=np.float32):
    df["responsibility"] = df.apply(
        lambda row: _to_numpy(
            row["responsibility"], literal_eval(row["responsibility_shape"]), dtype
        ),
        axis=1,
    )


def process_mask_column(df, column):
    df[column] = df.apply(
        lambda row: _to_numpy(
            row[column],
            literal_eval(row["mask_shape"]),
            np.bool_,
        ),
        axis=1,
    )


def db_to_pandas(db, dtype=np.float32, table="rex", process=True):
    """for interactive use"""
    df = _dataframe(db, table=table)

    if process:
        process_responsibility_column(df, dtype=dtype)
        # df["responsibility"] = df.apply(
        #     lambda row: _to_numpy(
        #         row["responsibility"], literal_eval(row["responsibility_shape"]), dtype
        #     ),
        #     axis=1,
        # )

        process_mask_column(df, "sufficiency_mask")
        # df["sufficiency_mask"] = df.apply(
        #     lambda row: _to_numpy(
        #         row["sufficiency_mask"],
        #         literal_eval(row["mask_shape"]),
        #         np.bool_,
        #     ),
        #     axis=1,
        # )

    return df


def update_database(
    db,
    explanation: Explanation | MultiExplanation,  # type: ignore
    time_taken: float,
    multi=False,
    clauses=None,
    analysis_results=None,
):
    target_map = explanation.target_map

    if isinstance(target_map, tt.Tensor):
        target_map = target_map.detach().cpu().numpy()

    target = explanation.data.target
    if target is None:
        logger.warning("unable to update database as target is None")
        return
    classification = int(target.classification)  # type: ignore

    # potentially enmpty fields in the database
    necessity_mask = None
    complete_mask = None
    inverse_classification = None
    inverse_confidence = None
    necessity_confidence = None
    complete_classification = None
    complete_confidence = None
    area = None
    entropy = None
    insertion_curve = None
    deletion_curve = None

    if not multi:
        if explanation.sufficiency_mask is None:
            logger.warning("unable to update database as explanation is empty")
            return

        sufficiency_mask = try_detach(explanation.sufficiency_mask)

        if explanation.necessity_mask is not None:
            necessity_mask = try_detach(explanation.necessity_mask)
            necessity_confidence = explanation.necessity_confidence  # type: ignore
            inverse_classification = explanation.contrastive_classification
            inverse_confidence = explanation.contrastive_confidence
        if hasattr(explanation, "complete_mask"):
            if explanation.complete_mask is None:
                complete_mask = None
                complete_confidence = None
                complete_classification = None
            else:
                complete_mask = try_detach(explanation.complete_mask)
                complete_confidence = explanation.completeness_confidence
                complete_classification = explanation.completeness_classification

        explanation_confidence = explanation.sufficiency_confidence

        if analysis_results is not None:
            area = analysis_results["area"]
            entropy = analysis_results["entropy"]
            insertion_curve = analysis_results["insertion_curve"]
            deletion_curve = analysis_results["deletion_curve"]

        add_to_database(
            db,
            explanation.args,
            classification,
            target.confidence,
            target_map,
            sufficiency_mask,
            explanation_confidence,
            necessity_mask,
            necessity_confidence,
            inverse_classification,
            inverse_confidence,
            complete_mask,
            complete_classification,
            complete_confidence,
            area,
            entropy,
            insertion_curve,
            deletion_curve,
            time_taken,
            int(explanation.run_stats["total_passing"]),
            int(explanation.run_stats["total_failing"]),
            int(explanation.run_stats["max_depth_reached"]),
            explanation.run_stats["avg_box_size"],
        )

    else:
        if type(explanation) is not MultiExplanation:
            logger.warning(
                "unable to update database, multi=True is only valid for MultiExplanation objects"
            )
            return
        else:
            if clauses is None:
                clauses = [i for i in range(0, len(explanation.explanations))]
            for c, sufficiency_mask in enumerate(explanation.explanations):
                if c in clauses:
                    sufficiency_mask = try_detach(explanation.explanations[c])
                    add_to_database(
                        db,
                        explanation.args,
                        classification,
                        target.confidence,
                        target_map,
                        sufficiency_mask,
                        explanation.explanation_confidences[c],
                        necessity_mask,
                        necessity_confidence,
                        inverse_classification,
                        inverse_confidence,
                        complete_mask,
                        complete_classification,
                        complete_confidence,
                        area,
                        entropy,
                        insertion_curve,
                        deletion_curve,
                        time_taken,
                        int(explanation.run_stats["total_passing"]),
                        int(explanation.run_stats["total_failing"]),
                        int(explanation.run_stats["max_depth_reached"]),
                        explanation.run_stats["avg_box_size"],
                        multi_no=c,
                    )

                else:
                    logger.info("not adding %s into the database", c)


def add_to_database(
    db,
    args: CausalArgs,
    target: int,
    confidence: float | None,
    responsibility,
    sufficiency_mask,
    sufficiency_confidence: float | None,
    contrastive_mask,
    contrastive_confidence: float | None,
    inverse_classification: int | None,
    inverse_confidence: float | None,
    complete_mask,
    complete_classification: int | None,
    complete_confidence,
    area: float | None,
    entropy: float | None,
    insertion_curve: float | None,
    deletion_curve: float | None,
    time_taken: float,
    passing: int,
    failing: int,
    depth_reached: int,
    avg_box_size: float,
    multi=False,
    multi_no=None,
):
    if multi:
        id = hash(str(datetime.now().time()) + str(multi_no))
    else:
        id = hash(str(datetime.now().time()))

    responsibility_shape = responsibility.shape
    explanation_shape = sufficiency_mask.shape

    object = DataBaseEntry(
        id,
        args.path,
        target,
        confidence,
        responsibility,
        responsibility_shape,
        sufficiency_mask,
        explanation_shape,
        sufficiency_confidence,
        time_taken,
        area=area,
        entropy=entropy,
        insertion_curve=insertion_curve,
        deletion_curve=deletion_curve,
        passing=passing,
        failing=failing,
        contrastive_mask=contrastive_mask,
        contrastive_confidence=contrastive_confidence,
        complete_mask=complete_mask,
        complete_classification=complete_classification,
        complete_confidence=complete_confidence,
        inverse_classification=inverse_classification,
        inverse_confidence=inverse_confidence,
        depth_reached=depth_reached,
        avg_box_size=avg_box_size,
        tree_depth=args.tree_depth,
        search_limit=args.search_limit,
        iters=args.iters,
        min_size=args.min_box_size,
        distribution=str(args.distribution),
        distribution_args=str(args.distribution_args),
    )
    # if object is not None:
    object.multi = multi
    object.multi_no = multi_no
    object.passing = passing
    object.failing = failing
    object.total_work = passing + failing
    object.method = str(args.strategy)
    if args.strategy == Strategy.Spatial:
        object.spatial_radius = args.spatial_initial_radius
        object.spatial_eta = args.spatial_radius_eta
    if args.strategy == Strategy.MultiSpotlight:
        object.spotlights = args.spotlights
        object.spotlight_size = args.spotlight_size
        object.spotlight_eta = args.spotlight_eta
        object.obj_function = args.spotlight_objective_function

    db.add(object)
    db.commit()


class Base(DeclarativeBase):
    pass


class NumpyType(sa.types.TypeDecorator):
    impl = sa.types.LargeBinary

    cache_ok = True

    def process_bind_param(self, value, dialect):
        if value is not None:
            return zlib.compress(value, 9)

    def process_result_value(self, value, dialect):
        return value


class DataBaseEntry(Base):
    __tablename__ = "rex"
    id = Column(Integer, primary_key=True)
    path = Column(Unicode(100))
    target = Column(Integer)
    confidence = Column(Float)
    time = Column(Float)
    responsibility = Column(NumpyType)
    responsibility_shape = Column(Unicode)
    total_work = Column(Integer)
    passing = Column(Integer)
    failing = Column(Integer)
    # basic explanation type
    sufficiency_mask = Column(NumpyType)
    mask_shape = Column(Unicode)
    sufficiency_confidence = Column(Float)
    # sufficient and necessary mask
    contrastive_mask = Column(NumpyType)
    contrastive_confidence = Column(Float)

    inverse_classification = Column(Integer)
    inverse_confidence = Column(Float)

    # complete mask
    complete_mask = Column(NumpyType)
    complete_classification = Column(Integer)
    complete_confidence = Column(Float)

    # analysis results, if available
    area = Column(Float)
    entropy = Column(Float)
    insertion_curve = Column(Float)
    deletion_curve = Column(Float)

    multi = Column(Boolean)
    multi_no = Column(Integer)

    # causal specific columns
    depth_reached = Column(Integer)
    avg_box_size = Column(Float)
    tree_depth = Column(Integer)
    search_limit_per_iter = Column(Integer)
    iters = Column(Integer)
    min_size = Column(Integer)
    distribution = Column(String)
    distribution_args = Column(String)

    # explanation specific columns
    spatial_radius = Column(Integer)
    spatial_eta = Column(Float)

    # spotlight columns
    method = Column(String)
    spotlights = Column(Integer)
    spotlight_size = Column(Integer)
    spotlight_eta = Column(Float)
    obj_function = Column(String)

    def __init__(
        self,
        id,
        path,
        target,
        confidence,
        responsibility,
        responsibility_shape,
        sufficiency_mask,
        mask_shape,
        sufficiency_confidence,
        time_taken,
        contrastive_mask=None,
        contrastive_confidence=None,
        inverse_classification=None,
        inverse_confidence=None,
        complete_mask=None,
        complete_classification=None,
        complete_confidence=None,
        passing=None,
        failing=None,
        area=None,
        entropy=None,
        insertion_curve=None,
        deletion_curve=None,
        total_work=None,
        multi=False,
        multi_no=None,
        depth_reached=None,
        avg_box_size=None,
        tree_depth=None,
        search_limit=None,
        iters=None,
        min_size=None,
        distribution=None,
        distribution_args=None,
        initial_radius=None,
        radius_eta=None,
        method=None,
        spotlights=0,
        spotlight_size=0,
        spotlight_eta=0.0,
        obj_function=None,
    ):
        self.id = id
        self.path = path
        self.target = target
        self.confidence = confidence
        self.responsibility = responsibility
        self.responsibility_shape = str(responsibility_shape)
        self.sufficiency_mask = sufficiency_mask
        self.mask_shape = str(mask_shape)
        self.sufficiency_confidence = sufficiency_confidence
        self.time = time_taken
        self.total_work = total_work
        self.passing = passing
        self.failing = failing

        # contrastive and complete explanations
        self.contrastive_mask = contrastive_mask
        self.contrastive_confidence = contrastive_confidence
        self.inverse_classification = inverse_classification
        self.inverse_confidence = inverse_confidence
        self.complete_mask = complete_mask
        self.complete_classification = complete_classification
        self.complete_confidence = complete_confidence

        # analysis results
        self.area = area
        self.entropy = entropy
        self.insertion_curve = insertion_curve
        self.deletion_curve = deletion_curve

        # multi status
        self.multi = multi
        self.multi_no = multi_no

        # causal
        self.depth_reached = depth_reached
        self.avg_box_size = avg_box_size
        self.tree_depth = tree_depth
        self.search_limit = search_limit
        self.iters = iters
        self.min_size = min_size
        self.distribution = distribution
        self.distribution_args = distribution_args
        # spatial
        self.spatial_radius = initial_radius
        self.spatial_eta = radius_eta
        self.method = method
        # spotlights
        self.spotlights = spotlights
        self.spotlight_size = spotlight_size
        self.spotlight_eta = spotlight_eta
        self.obj_function = obj_function


def initialise_rex_db(name, echo=False):
    engine = create_engine(f"sqlite:///{name}", echo=echo)
    Base.metadata.create_all(engine, tables=[DataBaseEntry.__table__], checkfirst=True)  # type: ignore
    Session = sessionmaker(bind=engine)
    s = Session()
    return s



#!/usr/bin/env python
from typing import Tuple

import numpy as np
import torch as tt
from scipy.integrate import simpson
from scipy.signal import periodogram
from scipy.stats import entropy

from rex_xai.explanation.explanation import Explanation
from rex_xai.mutants.mutant import _apply_to_data
from rex_xai.utils._utils import (
    get_map_locations,
    set_boolean_mask_value,
    try_detach,
    xlogx,
)


class Evaluation:
    # TODO does this need to be an object? Probably not...
    # TODO consider inheritance from Explanation object
    def __init__(self, explanation: Explanation) -> None:
        self.explanation = explanation

    def ratio(self) -> float:
        """Returns percentage of data required for sufficient explanation"""
        if self.explanation.necessity_mask is not None:
            mask = try_detach(self.explanation.necessity_mask)
        else:
            mask = try_detach(self.explanation.sufficiency_mask)

        try:
            return (
                tt.count_nonzero(mask)  # type: ignore
                / mask.size  # type: ignore
            ).item()
        except TypeError:
            return (
                np.count_nonzero(mask)  # type: ignore
                / mask.size  # type: ignore
            )

    def spectral_entropy(self) -> Tuple[float, float]:
        """
        This code is a simplified version of
        https://github.com/raphaelvallat/antropy/blob/master/src/antropy/entropy.py
        """
        _, psd = periodogram(self.explanation.target_map)
        psd_norm = psd / psd.sum()
        ent = -np.sum(xlogx(psd_norm))
        if len(psd_norm.shape) == 2:
            max_ent = np.log2(len(psd_norm[0]))
        else:
            max_ent = np.log2(len(psd_norm))
        return ent, max_ent

    def responsibility_entropy(self):
        flat_map = try_detach(self.explanation.target_map).ravel()
        uniform = np.ones(flat_map.shape)

        return entropy(flat_map, uniform, base=2)

    def robustness(self, lower=None, upper=None, repeats=2):
        if lower is None:
            lower = tt.min(self.explanation.data.data).item()  # type: ignore
        if upper is None:
            upper = tt.max(self.explanation.data.data).item()  # type: ignore

        mask_shape = self.explanation.sufficiency_mask.shape  # type: ignore

        robustness_shape = (self.explanation.args.batch_size,) + mask_shape

        good = 0
        bad = 0

        if self.explanation.necessity_mask is not None:
            mask = self.explanation.necessity_mask
        else:
            mask = self.explanation.sufficiency_mask

        for _ in (0, repeats):
            test_tensor = (
                tt.FloatTensor(*robustness_shape)
                .uniform_(lower, upper)
                .to(self.explanation.data.device)
            )

            assert mask is not None
            for j in range(0, self.explanation.args.batch_size):
                test_tensor[j] = tt.where(
                    mask,
                    self.explanation.data.data,  # type: ignore
                    test_tensor[j],
                )

            result = self.explanation.prediction_func(test_tensor)

            for p in result:
                if p.classification == self.explanation.data.target.classification:  # type: ignore
                    good += 1
                else:
                    bad += 1

        return (good, bad)

    def insertion_deletion_curve(self, prediction_func, normalise=False):
        # keep pyright happy...
        assert self.explanation.data.target is not None
        assert self.explanation.data.data is not None
        assert self.explanation.data.target.confidence is not None

        step = self.explanation.args.insertion_step
        ranking = get_map_locations(map=self.explanation.target_map)

        # initialise empty insertion and delection curves arrays
        insertion_curve = np.zeros(len(ranking) // step)
        deletion_curve = np.zeros(len(ranking) // step)

        # blank insertion mask
        insertion_mask = tt.zeros(
            self.explanation.data.data.squeeze(0).shape, dtype=tt.bool
        ).to(self.explanation.data.device)
        # dense deletion mask
        deletion_mask = tt.ones(
            self.explanation.data.data.squeeze(0).shape, dtype=tt.bool
        ).to(self.explanation.data.device)

        # edit the model shape to include a batch number, if necessary and make it a tuple
        model_shape = self.explanation.data.model_shape
        model_shape[0] = self.explanation.args.batch_size
        model_shape = tuple(model_shape)

        # initialise tensors for insertion and deletion mutants
        insertion_mutants = tt.empty(model_shape, dtype=tt.float32)
        deletion_mutants = tt.empty(model_shape, dtype=tt.float32)

        pointer = 0
        j = 0
        for i in range(0, len(ranking), step):
            chunk = ranking[i : i + step]
            for _, loc in chunk:
                # set insertion_mask values to true
                set_boolean_mask_value(
                    insertion_mask,
                    self.explanation.data.mode,
                    self.explanation.data.model_order,
                    loc,
                )
                # set deletion mask values to false
                set_boolean_mask_value(
                    deletion_mask,
                    self.explanation.data.mode,
                    self.explanation.data.model_order,
                    loc,
                    val=False,
                )
            insertion_mutants[j] = _apply_to_data(
                insertion_mask, self.explanation.data
            ).squeeze(0)
            deletion_mutants[j] = _apply_to_data(
                deletion_mask, self.explanation.data
            ).squeeze(0)
            j += 1

            if j >= self.explanation.args.batch_size - 1:
                insertion_update, deletion_update = self.__batch(
                    insertion_mutants, deletion_mutants, prediction_func
                )
                insertion_curve[pointer : pointer + len(insertion_update)] = (
                    insertion_update
                )
                deletion_curve[pointer : pointer + len(insertion_update)] = (
                    deletion_update
                )

                pointer += j
                j = 0
                insertion_mutants = tt.empty(model_shape, dtype=tt.float32)
                deletion_mutants = tt.empty(model_shape, dtype=tt.float32)

        remaining = len(insertion_curve) - pointer
        insertion_update, deletion_update = self.__batch(
            insertion_mutants, deletion_mutants, prediction_func
        )
        insertion_curve[pointer : pointer + remaining] = insertion_update[:remaining]
        deletion_curve[pointer : pointer + remaining] = deletion_update[:remaining]

        i_auc = simpson(insertion_curve, dx=step)
        d_auc = simpson(deletion_curve, dx=step)

        if normalise:
            const = self.explanation.data.target.confidence * len(ranking)
            i_auc /= const
            d_auc /= const

        return i_auc, d_auc

    def __batch(
        self,
        im,
        dm,
        prediction_func,
    ):
        ip = prediction_func(im.to(self.explanation.data.device), raw=True)
        ipe = [
            ip[p, self.explanation.data.target.classification].item()  # type: ignore
            for p in range(0, ip.shape[0])
        ]
        dp = prediction_func(dm.to(self.explanation.data.device), raw=True)
        dpe = [
            dp[p, self.explanation.data.target.classification].item()  # type: ignore
            for p in range(0, dm.shape[0])
        ]
        return ipe, dpe



#!/usr/bin/env python3

import importlib.metadata
from enum import Enum
from itertools import chain, combinations
from typing import Dict, List, Tuple, Union

import numpy as np
import torch as tt
from numpy.typing import NDArray
from skimage.segmentation import mark_boundaries

from rex_xai.mutants.box import Box
from rex_xai.responsibility.prediction import Prediction
from rex_xai.utils.logger import logger

Strategy = Enum("Strategy", ["Global", "Spatial", "MultiSpotlight", "Contrastive"])

Queue = Enum("Queue", ["Area", "All", "Intersection", "DC"])

SpatialSearch = Enum("SpatialSearch", ["NotFound", "Found"])

ResponsibilityStyle = Enum("ResponsibilityStyle", ["Additive", "Multiplicative"])


def update_mask_shape(batch_size: int, mask_shape) -> Tuple[int]:
    if isinstance(mask_shape, tuple):
        mask_shape = list(mask_shape)

    mask_shape[0] = batch_size
    return tuple(mask_shape)


class ReXPositions:
    def __init__(
        self,
        sufficient_position=None,
        contrastive_position=None,
        sufficiency_found=False,
        current_max_necessity=0.0,
    ) -> None:
        self.sufficient_position = sufficient_position
        self.contrastive_position = contrastive_position
        self.sufficiency_found = sufficiency_found
        self.current_max_necessity = current_max_necessity

    def __repr__(self) -> str:
        return f"suff: {self.sufficient_position}, con: {self.contrastive_position}, suff_found: {self.sufficiency_found}"

    def is_empty(self):
        return self.sufficient_position is None and self.contrastive_position is None


def try_rounding(n, rounding: int | None) -> float:
    if rounding is None:
        return n
    else:
        return round(n, rounding)


def find_required_prediction(
    target: int,
    threshold: float,
    insertion_predictions: List[Prediction],
    contrastive_completeness_threshold: float = 0.0,
    deletion_predictions: List[Prediction] | None = None,
    rounding=None,
    sufficiency_found=False,
):
    positions = ReXPositions(sufficiency_found=sufficiency_found)
    if deletion_predictions is None:
        for i, p in enumerate(insertion_predictions):
            local_confidence = try_rounding(p.confidence, rounding)
            threshold = try_rounding(threshold, rounding)
            if p.classification == target and local_confidence >= threshold:  # type: ignore
                positions.sufficient_position = i
                return positions
    else:
        for i in range(0, len(insertion_predictions)):
            local_confidence = try_rounding(
                insertion_predictions[i].confidence, rounding
            )
            threshold = try_rounding(threshold, rounding)
            contrastive_completeness_threshold = try_rounding(
                contrastive_completeness_threshold, rounding
            )

            # check for a sufficiency
            if (
                insertion_predictions[i].classification == target
                and local_confidence >= threshold  # type: ignore
                and not positions.sufficiency_found
            ):
                positions.sufficient_position = i
                positions.sufficiency_found = True

            # check for necessity above threshold
            if (
                insertion_predictions[i].classification == target
                and local_confidence >= contrastive_completeness_threshold
                and deletion_predictions[i].classification != target
            ):
                positions.contrastive_position = i
    return positions


def find_complete_prediction(
    target: int,
    target_confidence,
    insertion_predictions: List[Prediction],
    rounding=2,
):
    for i in range(0, len(insertion_predictions)):
        p = insertion_predictions[i]
        if (
            p.classification == target
            and round(p.confidence, rounding) == round(target_confidence, rounding)  # type: ignore
        ):
            return i
    return None


def match_resposnibility_style(s: str) -> ResponsibilityStyle:
    if s == "additive":
        return ResponsibilityStyle.Additive
    if s == "multiplicative":
        return ResponsibilityStyle.Multiplicative
    raise ReXTomlError(f"{s} is an unknown responsibility style")


def try_detach(t) -> np.ndarray:
    if isinstance(t, tt.Tensor):
        return t.detach().cpu().numpy()
    elif isinstance(t, np.ndarray):
        return t
    else:
        raise ReXDataError("trying to convert a non-array into a numpy array")


def one_d_permute(tensor):
    perm = tt.randperm(len(tensor))
    return tensor[perm], perm


def powerset(r, reverse=True):
    ps = list(chain.from_iterable(combinations(r, lim) for lim in range(1, len(r) + 1)))
    if reverse:
        return reversed(ps)
    else:
        return ps


def clause_area(clause, areas: Dict) -> int:
    tot = 0
    for c in clause:
        tot += areas[c]
    return tot


class ReXError(Exception):
    pass


class ReXTomlError(ReXError):
    def __init__(self, message) -> None:
        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        return f"ReXTomlError: {self.message}"


class ReXPathError(ReXError):
    def __init__(self, message) -> None:
        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        return f"ReXPathError: no such file exists at {self.message}"


class ReXScriptError(ReXError):
    def __init__(self, message) -> None:
        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        return f"ReXScriptError: {self.message}"


class ReXDataError(ReXError):
    def __init__(self, message) -> None:
        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        return f"ReXDataError: {self.message}"


class ReXMapError(ReXError):
    def __init__(self, message) -> None:
        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        return f"ReXMapError: {self.message}"


def xlogx(ps):
    f = np.vectorize(_xlogx)
    return f(ps)


def _xlogx(p):
    if p == 0.0:
        return 0.0
    else:
        return p * np.log2(p)


def add_boundaries(
    img: Union[NDArray, tt.Tensor], segs: NDArray, colour=None
) -> NDArray:
    if colour is None:
        m = mark_boundaries(img, segs, mode="inner")
    else:
        m = mark_boundaries(img, segs, colour, mode="inner")
    m *= 255  # type: ignore
    m = m.astype(np.uint8)
    return m


def get_device(gpu: bool):
    if tt.backends.mps.is_available() and gpu:
        return tt.device("mps")
    if tt.device("cuda") and gpu:
        return tt.device("cuda")
    if gpu:
        logger.warning("gpu not available")
    return tt.device("cpu")


def get_map_locations(map, reverse=True):
    if isinstance(map, tt.Tensor):
        map = map.detach().cpu().numpy()
    coords = []
    for i, r in enumerate(np.nditer(map)):
        coords.append((r, np.unravel_index(i, map.shape)))
    coords = sorted(coords, reverse=reverse)
    return coords


def set_boolean_mask_value(
    tensor,
    mode,
    order,
    coords: Union[Box, Tuple[NDArray, NDArray]],
    val: bool = True,
):
    if isinstance(coords, Box):
        if mode in ("spectral", "tabular"):
            h = coords.col_start
            w = coords.col_stop
        elif mode == "voxel":
            h = slice(coords.row_start, coords.row_stop)
            w = slice(coords.col_start, coords.col_stop)
            d = slice(coords.depth_start, coords.depth_stop)
        else:
            h = slice(coords.row_start, coords.row_stop)
            w = slice(coords.col_start, coords.col_stop)
    else:
        if mode == "voxel":
            h = coords[0]
            w = coords[1]
            d = coords[2]  # type: ignore
        else:
            h = coords[0]
            w = coords[1]

    # three channels
    if mode == "RGB":
        # (C, H, W)
        if order == "first":
            tensor[:, h, w] = val
        # (H, W, C)
        else:
            tensor[h, w, :] = val
    elif mode in ("spectral", "tabular"):
        if len(tensor.shape) == 1:
            tensor[h:w] = val
        else:
            tensor[0, h:w] = val
    elif mode == "voxel":
        tensor[h, w, d] = val  # type: ignore
    else:
        raise ReXError("mode not recognised")


def ff(obj, fmt):
    """
    Like format(obj, fmt), but returns the string 'None' if obj is None.
    See the help for format() to see acceptable values for fmt.
    """
    return "None" if obj is None else format(obj, fmt)


def version():
    return importlib.metadata.version("rex-xai")



#!/usr/bin/env python
from __future__ import annotations

import re
from typing import Dict

import torch as tt
from tqdm import tqdm

from rex_xai.input.config import CausalArgs, Strategy
from rex_xai.input.input_data import Data
from rex_xai.mutants.mutant import _apply_to_data
from rex_xai.output import visualisation
from rex_xai.responsibility.resp_maps import ResponsibilityMaps
from rex_xai.utils._utils import (
    ReXPositions,
    SpatialSearch,
    find_complete_prediction,
    find_required_prediction,
    get_map_locations,
    set_boolean_mask_value,
    update_mask_shape,
)
from rex_xai.utils.logger import logger


class Explanation:
    def __init__(
        self,
        maps: ResponsibilityMaps,
        prediction_func,
        data: Data,
        args: CausalArgs,
        run_stats: Dict[str, float],
        keep_all_maps=False,
    ) -> None:
        if data.target is None or data.target.classification is None:
            raise (
                ValueError(
                    "Data must have `target` defined to create an Explanation object!"
                )
            )

        if keep_all_maps:
            self.maps = maps
        else:
            maps.subset(data.target.classification)
            self.maps = maps

        self.target_map: tt.Tensor | None = tt.from_numpy(
            maps.get(data.target.classification)
        ).to(data.device)
        if self.target_map is None:
            raise ValueError(
                f"No responsibility map found for target {data.target.classification}!"
            )

        self.sufficiency_mask: tt.Tensor | None = None
        self.sufficiency_confidence: float | None = None
        self.necessity_mask: tt.Tensor | None = None
        self.complete_mask: tt.Tensor | None = None
        self.prediction_func = prediction_func
        self.data: Data = data
        self.args: CausalArgs = args
        self.run_stats: Dict[str, float] = run_stats

    def __repr__(self) -> str:
        pred_func = repr(self.prediction_func)
        match_func_name = re.search(r"(<function .+) at", pred_func)
        if match_func_name:
            pred_func = match_func_name.group(1) + " >"

        run_stats = {k: round(v, 5) for k, v in self.run_stats.items()}

        exp_text = (
            "Explanation:"
            + f"\n\tCausalArgs: {type(self.args)}"
            + f"\n\tData: {self.data}"
            + f"\n\tprediction function: {pred_func}"
            + f"\n\tResponsibilityMaps: {self.maps}"
            + f"\n\trun statistics: {run_stats} (5 dp)"
        )

        return (
            exp_text
            + f"\n\tsufficiency mask: {self.sufficiency_mask}"
            + f"\n\texplanation confidence: {self.sufficiency_confidence:.4f}"
        )

    def extract(self):
        self.blank()
        if self.args.strategy == Strategy.Global:
            self.__global()
        if self.args.strategy == Strategy.Contrastive:
            self.contrastive()
        if self.args.strategy == Strategy.Spatial:
            if self.data.mode == "spectral":
                logger.warning(
                    "spatial search not yet implemented for spectral data, so defaulting to global search"
                )
                _ = self.__global()
            else:
                _ = self.__spatial()

    def blank(self):
        assert self.data.data is not None
        self.sufficiency_mask = tt.zeros(
            self.data.data.shape, dtype=tt.bool, device=self.data.device
        )

    def set_to_true(self, coords, mask=None):
        if mask is not None:
            mask = set_boolean_mask_value(
                mask, self.data.mode, self.data.model_order, coords
            )

    def __build_insertion_mask(self, ranking, chunk_pointer, ind, mask, mask_memo):
        chunk = ranking[chunk_pointer : chunk_pointer + self.args.chunk_size]

        if chunk == []:
            return ind, chunk_pointer, True

        # if the chunk is not empty...
        for _, loc in chunk:
            set_boolean_mask_value(
                mask[ind],
                self.data.mode,
                self.data.model_order,
                loc,
            )

        if ind == 0 and mask_memo is not None:
            mask[ind] = tt.logical_or(mask_memo, mask[ind])
        if ind > 0:
            mask[ind] = tt.logical_or(mask[ind - 1], mask[ind])

        chunk_pointer += self.args.chunk_size
        ind += 1
        return ind, chunk_pointer, False

    def __global(self, map=None, rounding=4):
        # get responsibility map and ranking
        if map is None:
            map = self.target_map
        ranking = get_map_locations(map)

        mask_shape = update_mask_shape(self.args.batch_size, self.data.model_shape)
        insertion_mask = tt.zeros(mask_shape, dtype=tt.bool).to(self.data.device)
        insertion_memo = None

        target_confidence: float = round(
            self.data.target.confidence * self.args.minimum_confidence_threshold,  # type: ignore
            rounding,
        )
        chunk_pointer = 0
        ind = 0
        sufficient_found = False

        # main loop
        with tqdm(
            total=len(ranking) // self.args.chunk_size,
            desc="Extracting global explanation",
        ) as pbar:
            while not sufficient_found:
                ind, chunk_pointer, exhausted = self.__build_insertion_mask(
                    ranking,
                    chunk_pointer,
                    ind,
                    insertion_mask,
                    insertion_memo,
                )

                pbar.update(1)

                if ind == self.args.batch_size or exhausted:
                    if exhausted:
                        insertion_mask = insertion_mask[:ind]

                    sufficient = self.prediction_func(
                        _apply_to_data(insertion_mask, self.data)
                    )

                    positions: ReXPositions = find_required_prediction(
                        self.data.target.classification,  # type: ignore
                        target_confidence,
                        sufficient,
                        rounding=rounding,
                    )

                    if not positions.is_empty():
                        if not sufficient_found:
                            self.sufficiency_mask = (
                                insertion_mask[positions.sufficient_position]
                                .detach()
                                .clone()
                            )
                            self.sufficiency_confidence = sufficient[
                                positions.sufficient_position
                            ].confidence
                            sufficient_found = True
                            logger.info(
                                f"a sufficient explanation for {self.data.target.classification} found with confidence {self.sufficiency_confidence:.4f}"
                            )

                            if False not in tt.unique(self.sufficiency_mask):
                                logger.info(
                                    f"the entire input was required to get sufficiency at {self.sufficiency_confidence:.4f} confidence",
                                )

                            return self.sufficiency_confidence
                    ind = 0
                    insertion_memo = insertion_mask[-1]

    def __generate_circle_coordinates(self, centre, radius: int):
        assert self.data.model_height is not None
        assert self.data.model_width is not None
        Y, X = tt.meshgrid(
            tt.arange(0, self.data.model_height),
            tt.arange(0, self.data.model_width),
            indexing="ij",
        )

        dist_from_centre = tt.sqrt(
            (Y.to(self.data.device) - centre[0]) ** 2
            + (X.to(self.data.device) - centre[1]) ** 2
        )

        # this produces a H * W mask which can be using in conjunction with tt.where()
        circle_mask = dist_from_centre <= radius

        return circle_mask

    def __draw_circle(self, centre, start_radius=None):
        if start_radius is None:
            start_radius = self.args.spatial_initial_radius
        mask = tt.zeros(
            self.data.model_shape[1:], dtype=tt.bool, device=self.data.device
        )
        circle_mask = self.__generate_circle_coordinates(centre, start_radius)
        if self.data.model_order == "first":
            mask[:, circle_mask] = True
        else:
            mask[circle_mask, :] = True
        return start_radius, circle_mask, mask

    def compute_masked_responsibility(self, mask):
        try:
            masked_responsibility = tt.where(
                mask,
                self.target_map,  # type: ignore
                self.data.mask_value,  # type: ignore
            )  # type: ignore
        except RuntimeError:
            masked_responsibility = tt.where(
                mask.permute((2, 0, 1)),
                self.target_map,  # type: ignore
                self.data.mask_value,  # type: ignore
            )  # type: ignore
        except Exception as e:
            logger.fatal(e)
            exit()

        logger.debug("using %s", self.args.spotlight_objective_function)
        if self.args.spotlight_objective_function == "mean":
            return tt.mean(masked_responsibility).item()
        if self.args.spotlight_objective_function == "max":
            return tt.max(masked_responsibility).item()

        logger.warning(
            "unable to understand %s, so using mean for search",
            self.args.spotlight_objective_function,
        )
        return tt.mean(masked_responsibility).item()

    def __spatial(self, centre=None, expansion_limit=None):
        # TODO  rewrite to use batching
        # we don't have a search location to start from, so we try to isolate one
        map = self.target_map
        if centre is None:
            centre = tt.unravel_index(tt.argmax(map), map.shape)  # type: ignore

        start_radius, circle, mask = self.__draw_circle(centre)

        if self.args.spotlight_objective_function == "none":
            masked_responsibility = None
        else:
            masked_responsibility = self.compute_masked_responsibility(mask)

        expansions = 0
        cutoff = (
            self.data.model_width * self.data.model_height * self.data.model_channels  # type: ignore
        )
        while tt.count_nonzero(mask) < cutoff:
            if expansion_limit is not None:
                if expansions >= expansion_limit and expansion_limit > 1:
                    logger.debug(
                        f"no explanation found after {expansion_limit} expansions"
                    )
                    return SpatialSearch.NotFound, masked_responsibility, None
            d = _apply_to_data(mask, self.data)
            p = self.prediction_func(d)[0]
            if (
                p.classification == self.data.target.classification  # type: ignore
                and p.confidence
                >= self.data.target.confidence * self.args.minimum_confidence_threshold  # type: ignore
            ):
                conf = self.__global(map=tt.where(circle, map, 0))  # type: ignore
                return SpatialSearch.Found, masked_responsibility, conf
            start_radius = int(start_radius * (1 + self.args.spatial_radius_eta))
            _, circle, _ = self.__draw_circle(centre, start_radius)
            if self.data.model_order == "first":
                mask[:, circle] = True
            else:
                mask[circle, :] = True
            expansions += 1

    def __build_contrastive_masks(
        self,
        ranking,
        ind,
        chunk_pointer,
        insertion_mask,
        insertion_memo,
        deletion_mask,
        deletion_memo,
    ):
        chunk = ranking[chunk_pointer : chunk_pointer + self.args.chunk_size]

        if chunk == []:
            return ind, chunk_pointer, True

        for _, loc in chunk:
            set_boolean_mask_value(
                insertion_mask[ind],
                self.data.mode,
                self.data.model_order,
                loc,
            )
            set_boolean_mask_value(
                deletion_mask[ind],
                self.data.mode,
                self.data.model_order,
                loc,
                val=False,
            )

        if ind == 0 and insertion_memo is not None and deletion_memo is not None:
            insertion_mask[ind] = tt.logical_or(insertion_memo, insertion_mask[ind])
            deletion_mask[ind] = tt.logical_and(deletion_memo, deletion_mask[ind])
        if ind > 0:
            insertion_mask[ind] = tt.logical_or(
                insertion_mask[ind - 1], insertion_mask[ind]
            )
            deletion_mask[ind] = tt.logical_and(
                deletion_mask[ind - 1], deletion_mask[ind]
            )

        chunk_pointer += self.args.chunk_size
        ind += 1

        return ind, chunk_pointer, False

    def __complete(
        self,
        ranking,
        insertion_mask,
        insertion_memo,
        mask_shape,
        starting_pointer,
        rounding=2,
    ):
        target_confidence = self.data.target.confidence  # type: ignore
        step = self.args.chunk_size

        # check that the confidences aren't already at the correct levels
        if round(self.necessity_confidence, rounding) == target_confidence:  # type: ignore
            logger.info("the sufficient and necessary explanation is already complete")
            self.complete_mask = None
            self.completeness_confidence = self.necessity_confidence
            return

        complete_explanation_found = False
        insertion_mask = insertion_mask.zero_()
        insertion_mask[0] = self.necessity_mask.detach().clone()  # type: ignore
        exhausted = False

        chunk_pointer = len(ranking)
        ind = 1

        with tqdm(
            total=(len(ranking) - starting_pointer) // self.args.chunk_size,
            desc="Completeness Explanation",
        ) as pbar:
            while not complete_explanation_found:
                chunk = ranking[chunk_pointer - step : chunk_pointer]

                if chunk == []:
                    logger.info("the entire image is required for completeness")
                    self.complete_mask = tt.logical_xor(
                        tt.ones(mask_shape[1:], dtype=tt.bool).to(self.data.device),
                        self.necessity_mask.detach().clone(),  # type: ignore
                    )
                    return

                for _, loc in chunk:
                    set_boolean_mask_value(
                        insertion_mask[ind],
                        self.data.mode,
                        self.data.model_order,
                        loc,
                    )

                pbar.update(1)
                if ind == 0 and insertion_memo is not None:
                    insertion_mask[ind] = tt.logical_or(
                        insertion_memo, insertion_mask[ind]
                    )
                else:
                    insertion_mask[ind] = tt.logical_or(
                        insertion_mask[ind - 1], insertion_mask[ind]
                    )

                ind += 1
                chunk_pointer -= step

                if chunk_pointer <= starting_pointer:
                    exhausted = True

                if ind == self.args.batch_size or exhausted:
                    if exhausted:
                        insertion_mask = insertion_mask[:ind]

                    complete_predictions = self.prediction_func(
                        _apply_to_data(insertion_mask, self.data)
                    )

                    position = find_complete_prediction(
                        self.data.target.classification,  # type: ignore
                        target_confidence,
                        complete_predictions,
                    )

                    if position is not None:
                        self.complete_mask = tt.logical_xor(
                            self.necessity_mask,  # type: ignore
                            insertion_mask[position],  # type: ignore
                        )
                        cp = self.prediction_func(
                            _apply_to_data(self.complete_mask, self.data)
                        )[0]
                        self.completeness_classification = cp.classification
                        self.completeness_confidence = cp.confidence
                        complete_explanation_found = True
                        logger.info(
                            "found complete explanation for class %d. The completeness mask is class %d with confidence %.2f of size %d",
                            self.data.target.classification,  # type: ignore
                            self.completeness_classification,
                            self.completeness_confidence,
                            tt.count_nonzero(self.complete_mask)  # type: ignore
                            // self.data.model_channels,
                        )
                    else:
                        ind = 0

    def contrastive(self):
        rounding = 4
        mask_shape = update_mask_shape(self.args.batch_size, self.data.model_shape)

        insertion_mask = tt.zeros(mask_shape, dtype=tt.bool).to(self.data.device)
        deletion_mask = tt.ones(mask_shape, dtype=tt.bool).to(self.data.device)

        ranking = get_map_locations(map=self.target_map)

        target_confidence: float = (
            self.args.minimum_confidence_threshold * self.data.target.confidence  # type: ignore
        )
        contrastive_completeness_threshold: float = self.data.target.confidence  # type: ignore

        self.sufficiency_confidence = None
        self.necessity_mask = None
        self.necessity_confidence = None
        # the class and confidence of the pixels after self.necessity_mask has been removed
        self.contrastive_classification = None
        self.contrastive_confidence = None

        sufficient_found = False
        contrastive_found = False

        insertion_memo = None
        deletion_memo = None
        chunk_pointer = 0
        ind = 0

        with tqdm(
            total=len(ranking) // self.args.batch_size,
            desc="Calculating Contrastive Explanation",
        ) as pbar:
            while not contrastive_found:
                ind, chunk_pointer, exhausted = self.__build_contrastive_masks(
                    ranking,
                    ind,
                    chunk_pointer,
                    insertion_mask,
                    insertion_memo,
                    deletion_mask,
                    deletion_memo,
                )

                pbar.update(1)
                # we have filled insertion_mask and deletion_mask with test cases, now time to test
                if ind == self.args.batch_size or exhausted:
                    if exhausted:
                        insertion_mask = insertion_mask[:ind]
                        deletion_mask = deletion_mask[:ind]

                    sufficient = self.prediction_func(
                        _apply_to_data(insertion_mask, self.data)
                    )
                    contrastive = self.prediction_func(
                        _apply_to_data(deletion_mask, self.data)
                    )

                    positions: ReXPositions = find_required_prediction(
                        self.data.target.classification,  # type: ignore
                        target_confidence,
                        sufficient,
                        contrastive_completeness_threshold,
                        contrastive,
                        rounding=rounding,
                        sufficiency_found=sufficient_found,
                    )

                    if positions.sufficient_position is not None:
                        if not sufficient_found:
                            sufficient_found = True
                            self.sufficiency_mask = (
                                insertion_mask[positions.sufficient_position]
                                .detach()
                                .clone()
                            )
                            self.sufficiency_confidence = sufficient[
                                positions.sufficient_position
                            ].confidence
                            logger.info(
                                "a sufficient explanation for %d found with confidence %.4f of size %d",
                                self.data.target.classification,  # type: ignore
                                self.sufficiency_confidence,
                                tt.count_nonzero(self.sufficiency_mask)  # type: ignore
                                // self.data.model_channels,
                            )

                            if False not in tt.unique(self.sufficiency_mask):
                                logger.info(
                                    "the entire input was required to get sufficiency at %.4f confidence, so there is no independent necessity or completeness",
                                    self.sufficiency_confidence,
                                )
                                self.args.strategy = Strategy.Global
                                self.args.complete = False
                                return

                    if positions.contrastive_position is not None:
                        contrastive_found = True
                        self.necessity_mask = (
                            insertion_mask[positions.contrastive_position]
                            .detach()
                            .clone()
                        )
                        self.necessity_confidence = sufficient[
                            positions.contrastive_position
                        ].confidence
                        self.contrastive_classification = contrastive[
                            positions.contrastive_position
                        ].classification
                        self.contrastive_confidence = contrastive[
                            positions.contrastive_position
                        ].confidence

                        logger.info(
                            "a contrastive explanation for %d (now %d) found with confidence %.3f of size %d",
                            self.data.target.classification,  # type: ignore
                            self.contrastive_classification,
                            self.necessity_confidence,
                            tt.count_nonzero(self.necessity_mask)  # type: ignore
                            // self.data.model_channels,
                        )

                        if tt.count_nonzero(self.sufficiency_mask) == tt.count_nonzero(  # type: ignore
                            self.necessity_mask
                        ):
                            logger.info(
                                "there is no difference between sufficiency and necessity on this input"
                            )

                        if False not in self.necessity_mask:
                            logger.info(
                                "the sufficient and necessery explanation is already complete"
                            )
                            self.args.complete = False

                        if self.args.complete:
                            return self.__complete(
                                ranking,
                                insertion_mask,
                                insertion_memo,
                                mask_shape,
                                chunk_pointer,
                            )
                        return

                    insertion_memo = insertion_mask[-1]
                    deletion_memo = deletion_mask[-1]
                    ind = 0

    def save(self, path, mask=None):
        if mask is not None:
            visualisation.save_image(
                mask,
                self.data,
                self.args,
                path=path,
            )
        assert self.sufficiency_mask is not None
        if self.data.mode in ("RGB", "voxel") and mask is None:
            if self.args.complete:
                visualisation.save_complete(self, self.data, self.args, path=path)

            elif self.args.strategy == Strategy.Contrastive:
                visualisation.save_contrastive(self, self.data, self.args, path=path)
            else:
                mask = (
                    self.sufficiency_mask
                    if self.necessity_mask is None
                    else self.necessity_mask
                )
                visualisation.save_image(
                    self.sufficiency_mask,
                    self.data,
                    self.args,
                    path=path,
                )

        if self.data.mode == "spectral":
            visualisation.spectral_plot(
                self.sufficiency_mask,
                self.data,
                self.target_map,
                self.args.heatmap_colours,
                path=path,
            )
        if self.data.mode == "tabular":
            pass

    def heatmap_plot(self, path=None):
        if self.target_map is not None:
            if self.data.mode == "RGB":
                visualisation.heatmap_plot(
                    self.data,
                    self.target_map,
                    self.args.heatmap_colours,
                    path=path,
                )
            elif self.data.mode == "voxel":
                visualisation.voxel_plot(
                    self.args,
                    self.target_map,  # type: ignore
                    self.data,
                    path=path,
                )
            else:
                return NotImplementedError

    def surface_plot(self, path=None):
        if self.data.mode == "RGB":
            visualisation.surface_plot(
                self.data.input,
                self.args,
                self.target_map,  # type: ignore
                self.data.target,  #  type: ignore
                path=path,
            )
        elif self.data.mode == "voxel":
            logger.warning(
                "Surface plot not available for voxel data using voxel plot instead"
            )
            visualisation.voxel_plot(
                self.args,
                self.target_map,  # type: ignore
                self.data,
                path=path,
            )
        else:
            return NotImplementedError

    def show(self, path=None):
        assert self.sufficiency_mask is not None
        if self.data.mode in ("RGB", "voxel"):
            out = visualisation.save_image(
                self.sufficiency_mask,
                self.data,
                self.args,
                path=path,
            )
            return out
        else:
            return NotImplementedError
